"""
This module provides utility functions for printing messages with different levels of importance, such as:
- info()
- debug()
- suggestion()
- progress()
- warning()
- error()

It also includes a function to print the type of each value and the value itself:
- whatisit()
"""

# Imports
import sys
import time
from typing import Callable, Any

# Colors constants
RESET: str   = "\033[0m"
RED: str     = "\033[91m"
GREEN: str   = "\033[92m"
YELLOW: str  = "\033[93m"
BLUE: str    = "\033[94m"
MAGENTA: str = "\033[95m"
CYAN: str    = "\033[96m"
LINE_UP: str = "\033[1A"

# Print functions
previous_args_kwards: tuple[tuple[Any, ...], dict[str, Any]] = ((), {})
nb_values: int = 1

def is_same_print(*args: tuple[Any, ...], **kwargs: dict[str, Any]) -> bool:
	""" Checks if the current print call is the same as the previous one.

	Args:
		args (tuple): The arguments passed to the print function.
		kwargs (dict): The keyword arguments passed to the print function.
	Returns:
		bool: True if the current print call is the same as the previous one, False otherwise.
	"""
	global previous_args_kwards, nb_values
	if previous_args_kwards == (args, kwargs):
		nb_values += 1
		return True
	else:
		previous_args_kwards = (args, kwargs)
		nb_values = 1
		return False

def current_time() -> str:
	""" Get the current time in the format HH:MM:SS	"""
	return time.strftime("%H:%M:%S")

def info(*values: Any, prefix: str = "", **print_kwargs: Any) -> None:
	""" Print an information message looking like "[INFO HH:MM:SS] message"

	Args:
		values			(Any):	Values to print (like the print function)
		prefix			(str):		Prefix to add to the values
		print_kwargs	(dict):		Keyword arguments to pass to the print function
	"""
	if not is_same_print(*values, **print_kwargs):
		print(f"{prefix}{GREEN}[INFO  {current_time()}]", *values, RESET, **print_kwargs)
	else:
		print(f"{LINE_UP}{prefix}{GREEN}[INFO  {current_time()}] (x{nb_values})", *values, RESET, **print_kwargs)

def debug(*values: Any, prefix: str = "", **print_kwargs: Any) -> None:
	""" Print a debug message looking like "[DEBUG HH:MM:SS] message"

	Args:
		values			(Any):		Values to print (like the print function)
		prefix			(str):		Prefix to add to the values
		print_kwargs	(dict):		Keyword arguments to pass to the print function
	"""
	if not is_same_print(*values, **print_kwargs):
		print(f"{prefix}{BLUE}[DEBUG {current_time()}]", *values, RESET, **print_kwargs)
	else:
		print(f"{LINE_UP}{prefix}{BLUE}[DEBUG {current_time()}] (x{nb_values})", *values, RESET, **print_kwargs)

def suggestion(*values: Any, prefix: str = "", **print_kwargs: Any) -> None:
	""" Print a suggestion message looking like "[SUGGESTION HH:MM:SS] message"

	Args:
		values			(Any):		Values to print (like the print function)
		prefix			(str):		Prefix to add to the values
		print_kwargs	(dict):		Keyword arguments to pass to the print function
	"""
	if not is_same_print(*values, **print_kwargs):
		print(f"{prefix}{CYAN}[SUGGESTION {current_time()}]", *values, RESET, **print_kwargs)
	else:
		print(f"{LINE_UP}{prefix}{CYAN}[SUGGESTION {current_time()}] (x{nb_values})", *values, RESET, **print_kwargs)

def progress(*values: Any, prefix: str = "", **print_kwargs: Any) -> None:
	""" Print a progress message looking like "[PROGRESS HH:MM:SS] message"

	Args:
		values			(Any):		Values to print (like the print function)
		prefix			(str):		Prefix to add to the values
		print_kwargs	(dict):		Keyword arguments to pass to the print function
	"""
	if not is_same_print(*values, **print_kwargs):
		print(f"{prefix}{MAGENTA}[PROGRESS {current_time()}]", *values, RESET, **print_kwargs)
	else:
		print(f"{LINE_UP}{prefix}{MAGENTA}[PROGRESS {current_time()}] (x{nb_values})", *values, RESET, **print_kwargs)

def warning(*values: Any, prefix: str = "", **print_kwargs: Any) -> None:
	""" Print a warning message looking like "[WARNING HH:MM:SS] message"

	Args:
		values			(Any):		Values to print (like the print function)
		prefix			(str):		Prefix to add to the values
		print_kwargs	(dict):		Keyword arguments to pass to the print function
	"""
	if not is_same_print(*values, **print_kwargs):
		print(f"{prefix}{YELLOW}[WARNING {current_time()}]", *values, RESET, **print_kwargs)
	else:
		print(f"{LINE_UP}{prefix}{YELLOW}[WARNING {current_time()}] (x{nb_values})", *values, RESET, **print_kwargs)

def error(*values: Any, exit: bool = True, prefix: str = "", **print_kwargs: Any) -> None:
	""" Print an error message and optionally ask the user to continue or stop the program

	Args:
		values			(Any):		Values to print (like the print function)
		exit			(bool):			Whether to ask the user to continue or stop the program, false to ignore the error automatically and continue
		print_kwargs	(dict):			Keyword arguments to pass to the print function
	"""
	if not is_same_print(*values, **print_kwargs):
		print(f"{prefix}{RED}[ERROR {current_time()}]", *values, RESET, **print_kwargs)
	else:
		print(f"{LINE_UP}{prefix}{RED}[ERROR {current_time()}] (x{nb_values})", *values, RESET, **print_kwargs)
	if exit:
		try:
			input("Press enter to ignore error and continue or 'CTRL+C' to stop the program... ")
		except KeyboardInterrupt:
			sys.exit(1)

def whatisit(*values: Any, print_function: Callable[..., None] = debug, prefix: str = "") -> None:
	""" Print the type of each value and the value itself

	Args:
		values			(Any):		Values to print
		print_function	(Callable):	Function to use to print the values
		prefix			(str):		Prefix to add to the values
	"""
	if len(values) > 1:
		print_function("(What is it?)", prefix=prefix)
		for value in values:
			print_function(f"{type(value)}:\t{value}", prefix=prefix)
	elif len(values) == 1:
		print_function(f"(What is it?) {type(values[0])}:\t{values[0]}", prefix=prefix)






if __name__ == "__main__":
	info("Hello", "World")
	time.sleep(1)
	info("Hello", "World")
	time.sleep(1)
	info("Hello", "World")
	time.sleep(1)
	info("Hello", "World")
	time.sleep(1)
	info("Hello", "World")
	time.sleep(1)
	info("Hello", "World")




"""
This module provides utility functions for parallel processing, such as:
- multiprocessing(): Execute a function in parallel using multiprocessing
- multithreading(): Execute a function in parallel using multithreading

I highly encourage you to read the function docstrings to understand when to use each method.
"""

# Imports
from .print import *
from .decorators import *
from multiprocessing import Pool, cpu_count
from typing import Callable, TypeVar
from tqdm import tqdm
from tqdm.contrib.concurrent import process_map # type: ignore
from concurrent.futures import ThreadPoolExecutor

# Small test functions for doctests
def doctest_square(x: int) -> int:
	return x * x
def doctest_multiply(args: tuple[int, int]) -> int:
	x, y = args
	return x * y
def doctest_slow(x: int) -> int:
	import time
	time.sleep(0.5)
	return x

# Constants
CPU_COUNT: int = cpu_count()
BAR_FORMAT: str = "{l_bar}{bar}" + MAGENTA + "| {n_fmt}/{total_fmt} [{rate_fmt}{postfix}, {elapsed}<{remaining}]" + RESET
T = TypeVar("T")
R = TypeVar("R")

@handle_error(error_log=LogLevels.ERROR_TRACEBACK)
def multiprocessing(func: Callable[[T], R], args: list[T], chunksize: int = 1, desc: str = "", max_workers: int = CPU_COUNT, verbose_depth: int = 0) -> list[R]:
	""" Method to execute a function in parallel using multiprocessing, you should use it:
	- For CPU-bound operations where the GIL (Global Interpreter Lock) is a bottleneck.
	- When the task can be divided into smaller, independent sub-tasks that can be executed concurrently.
	- For operations that involve heavy computations, such as scientific simulations, data processing, or machine learning tasks.

	Args:
		func			(Callable):			Function to execute
		args			(list):				List of arguments to pass to the function (no starmap will be used)
		chunksize		(int):				Number of arguments to process at a time (Defaults to 1 for proper progress bar display)
		desc			(str):				Description of the function execution displayed in the progress bar
		max_workers		(int):				Number of workers to use (Defaults to CPU_COUNT)
		verbose_depth	(int):				Level of verbosity, decrease by 1 for each depth
	Returns:
		list[object]:	Results of the function execution
	Examples:
		>>> multiprocessing(doctest_square, args=[1, 2, 3])
		[1, 4, 9]

		>>> multiprocessing(doctest_multiply, [(1,2), (3,4), (5,6)])
		[2, 12, 30]

		>>> # Will process in parallel with progress bar
		>>> multiprocessing(doctest_slow, list(range(10)), desc="Processing", verbose_depth=1)
		[0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
	"""
	if desc:
		desc = MAGENTA + desc

	# Do multiprocessing only if there is more than 1 argument and more than 1 CPU
	if max_workers > 1 and len(args) > 1:
		if verbose_depth > 0:
			return list(process_map(func, args, max_workers=max_workers, chunksize=chunksize, desc=desc, bar_format=BAR_FORMAT)) # type: ignore
		else:
			with Pool(max_workers) as pool:
				return list(pool.map(func, args, chunksize=chunksize))	# type: ignore

	# Single process execution
	else:
		if verbose_depth > 0:
			return [func(arg) for arg in tqdm(args, total=len(args), desc=desc, bar_format=BAR_FORMAT)]
		else:
			return [func(arg) for arg in args]


@handle_error(error_log=LogLevels.ERROR_TRACEBACK)
def multithreading(func: Callable[[T], R], args: list[T], desc: str = "", max_workers: int = CPU_COUNT, verbose_depth: int = 0) -> list[R]:
	""" Method to execute a function in parallel using multithreading, you should use it:
	- For I/O-bound operations where the GIL is not a bottleneck, such as network requests or disk operations.
	- When the task involves waiting for external resources, such as network responses or user input.
	- For operations that involve a lot of waiting, such as GUI event handling or handling user input.

	Args:
		func			(Callable):			Function to execute
		args			(list):				List of arguments to pass to the function (no starmap will be used)
		desc			(str):				Description of the function execution displayed in the progress bar
		max_workers		(int):				Number of workers to use (Defaults to CPU_COUNT)
		verbose_depth	(int):				Level of verbosity, decrease by 1 for each depth
	Returns:
		list[object]:	Results of the function execution
	Examples:
		>>> multithreading(doctest_square, args=[1, 2, 3])
		[1, 4, 9]

		>>> multithreading(doctest_multiply, [(1,2), (3,4), (5,6)])
		[2, 12, 30]

		>>> # Will process in parallel with progress bar
		>>> multithreading(doctest_slow, list(range(10)), desc="Threading", verbose_depth=1)
		[0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
	"""
	if desc:
		desc = MAGENTA + desc

	# Do multithreading only if there is more than 1 argument and more than 1 CPU
	if max_workers > 1 and len(args) > 1:
		if verbose_depth > 0:
			with ThreadPoolExecutor(max_workers) as executor:
				return list(tqdm(executor.map(func, args), total=len(args), desc=desc, bar_format=BAR_FORMAT))
		else:
			with ThreadPoolExecutor(max_workers) as executor:
				return list(executor.map(func, args))

	# Single process execution
	else:
		if verbose_depth > 0:
			return [func(arg) for arg in tqdm(args, total=len(args), desc=desc, bar_format=BAR_FORMAT)]
		else:
			return [func(arg) for arg in args]




"""
This module provides functions for creating and managing archives.
- make_archive: Make an archive with consistency using FILES_TO_WRITE variable
- repair_zip_file: Try to repair a corrupted zip file (NOT IMPLEMENTED)
"""

# Imports
from .io import *
from .print import *
from zipfile import ZipFile, ZipInfo, ZIP_DEFLATED

# Function that makes an archive with consistency (same zip file each time)
@handle_error()
def make_archive(
	source: str,
	destinations: list[str]|str = [],
	override_time: None | tuple[int, int, int, int, int, int] = None,
	create_dir: bool = False
) -> bool:
	""" Make an archive with consistency using FILES_TO_WRITE variable

	Args:
		source				(str):						The source folder to archive
		destinations		(list[str]|str):			The destination folder(s) or file(s) to copy the archive to
		override_time		(None | tuple[int, ...]):	The constant time to use for the archive (e.g. (2024, 1, 1, 0, 0, 0) for 2024-01-01 00:00:00)
		create_dir			(bool):						Whether to create the destination directory if it doesn't exist (default: False)
	Returns:
		bool: Always returns True unless any strong error
	"""
	# Fix copy_destinations type if needed
	if destinations and isinstance(destinations, str):
		destinations = [destinations]
	if not destinations:
		raise ValueError("destinations must be a list of at least one destination")

	# Create the archive
	destination: str = clean_path(destinations[0])
	destination = destination if ".zip" in destination else destination + ".zip"
	with ZipFile(destination, "w", compression=ZIP_DEFLATED, compresslevel=9) as zip:
		for root, _, files in os.walk(source):
			for file in files:
				file_path: str = clean_path(os.path.join(root, file))
				info: ZipInfo = ZipInfo(file_path)
				info.compress_type = ZIP_DEFLATED
				if override_time:
					info.date_time = override_time
				with open(file_path, "rb") as f:
					zip.writestr(info, f.read())

	# Copy the archive to the destination(s)
	for dest_folder in destinations:
		@handle_error(Exception, message=f"Unable to copy '{destination}' to '{dest_folder}'", error_log=LogLevels.WARNING)
		def internal(src: str, dest: str) -> None:
			super_copy(src, dest, create_dir=create_dir)
		internal(destination, clean_path(dest_folder))

	return True



# Function that repair a corrupted zip file (ignoring some of the errors)
from .dont_look.zip_file_override import ZipFileOverride
@handle_error()
def repair_zip_file(file_path: str, destination: str) -> bool:
	""" Try to repair a corrupted zip file by ignoring some of the errors

	Args:
		file_path		(str):	Path of the zip file to repair
		destination		(str):	Destination of the new file
	Returns:
		bool: Always returns True unless any strong error
	"""
	# Check
	if not os.path.exists(file_path):
		raise FileNotFoundError(f"File '{file_path}' not found")
	dirname: str = os.path.dirname(destination)
	if dirname and not os.path.exists(dirname):
		raise FileNotFoundError(f"Directory '{dirname}' not found")

	# Read it
	with ZipFileOverride(file_path, "r") as zip_file:

		# Get a list of all the files in the ZIP file
		file_list: list[str] = zip_file.namelist()

		# Create a new ZIP file at the destination
		with ZipFileOverride(destination, "w", compression=ZIP_DEFLATED) as new_zip_file:
			for file_name in file_list:
				try:
					new_zip_file.writestr(file_name, zip_file.read(file_name))
				except KeyboardInterrupt:
					continue
	
	return True





"""
This module provides context managers for temporarily silencing output.
- Muffle: Context manager that temporarily silences output (alternative to stouputils.decorators.silent())
"""

# Imports
import os
import sys
from typing import TextIO, Any

# Context manager to temporarily silence output
class Muffle:
	""" Context manager that temporarily silences output.

	Alternative to stouputils.decorators.silent()
	
	>>> with Muffle():
	...     print("This will not be printed")
	"""
	def __init__(self, mute_stderr: bool = False) -> None:
		self.mute_stderr: bool = mute_stderr

	def __enter__(self) -> None:
		self.original_stdout: TextIO = sys.stdout
		sys.stdout = open(os.devnull, 'w')
		if self.mute_stderr:
			self.original_stderr: TextIO = sys.stderr
			sys.stderr = open(os.devnull, 'w')

	def __exit__(self, exc_type: type[BaseException]|None, exc_val: BaseException|None, exc_tb: Any|None) -> None:
		sys.stdout.close()
		sys.stdout = self.original_stdout
		if self.mute_stderr:
			sys.stderr.close()
			sys.stderr = self.original_stderr


